"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NodeMessageHandler = void 0;
const core_1 = require("@zwave-js/core");
const error_1 = require("../error");
const state_1 = require("../state");
const command_1 = require("./command");
const __1 = require("..");
const common_1 = require("../common");
class NodeMessageHandler {
    async handle(message, clientsController, driver, client) {
        var _a, _b, _c, _d;
        const { nodeId, command } = message;
        const node = driver.controller.nodes.get(nodeId);
        if (!node) {
            throw new error_1.NodeNotFoundError(nodeId);
        }
        switch (message.command) {
            case command_1.NodeCommand.setValue: {
                const result = await node.setValue(message.valueId, message.value, message.options);
                return (0, common_1.setValueOutgoingMessage)(result, client.schemaVersion);
            }
            case command_1.NodeCommand.refreshInfo: {
                await node.refreshInfo(message.options);
                return {};
            }
            case command_1.NodeCommand.getDefinedValueIDs: {
                const valueIds = node.getDefinedValueIDs();
                return { valueIds };
            }
            case command_1.NodeCommand.getValueMetadata: {
                if (message.valueId.commandClass == core_1.CommandClasses.Configuration) {
                    return (0, state_1.dumpConfigurationMetadata)(node.getValueMetadata(message.valueId), client.schemaVersion);
                }
                return (0, state_1.dumpMetadata)(node.getValueMetadata(message.valueId), client.schemaVersion);
            }
            case command_1.NodeCommand.beginFirmwareUpdate: {
                const firmwareFile = Buffer.from(message.firmwareFile, "base64");
                let firmware = (0, core_1.extractFirmware)(firmwareFile, (_a = message.firmwareFileFormat) !== null && _a !== void 0 ? _a : (0, core_1.guessFirmwareFileFormat)(message.firmwareFilename, firmwareFile));
                // Defer to the target provided in the messaage when available
                firmware.firmwareTarget = (_b = message.target) !== null && _b !== void 0 ? _b : firmware.firmwareTarget;
                const result = await node.updateFirmware([firmware]);
                return (0, common_1.firmwareUpdateOutgoingMessage)(result, client.schemaVersion);
            }
            case command_1.NodeCommand.updateFirmware: {
                const updates = message.updates.map((update) => {
                    var _a, _b;
                    const file = Buffer.from(update.file, "base64");
                    let firmware = (0, core_1.extractFirmware)(file, (_a = update.fileFormat) !== null && _a !== void 0 ? _a : (0, core_1.guessFirmwareFileFormat)(update.filename, file));
                    // Defer to the target provided in the messaage when available
                    firmware.firmwareTarget =
                        (_b = update.firmwareTarget) !== null && _b !== void 0 ? _b : firmware.firmwareTarget;
                    return firmware;
                });
                const result = await node.updateFirmware(updates);
                return (0, common_1.firmwareUpdateOutgoingMessage)(result, client.schemaVersion);
            }
            case command_1.NodeCommand.abortFirmwareUpdate: {
                await node.abortFirmwareUpdate();
                return {};
            }
            case command_1.NodeCommand.getFirmwareUpdateCapabilities: {
                const capabilities = await node.getFirmwareUpdateCapabilities();
                return { capabilities };
            }
            case command_1.NodeCommand.getFirmwareUpdateCapabilitiesCached: {
                const capabilities = node.getFirmwareUpdateCapabilitiesCached();
                return { capabilities };
            }
            case command_1.NodeCommand.pollValue: {
                const value = await node.pollValue(message.valueId);
                return { value };
            }
            case command_1.NodeCommand.setRawConfigParameterValue: {
                await node.commandClasses.Configuration.set({
                    parameter: message.parameter,
                    value: message.value,
                    valueSize: message.valueSize,
                });
                return {};
            }
            case command_1.NodeCommand.refreshValues: {
                await node.refreshValues();
                return {};
            }
            case command_1.NodeCommand.refreshCCValues: {
                await node.refreshCCValues(message.commandClass);
                return {};
            }
            case command_1.NodeCommand.ping: {
                const responded = await node.ping();
                return { responded };
            }
            case command_1.NodeCommand.hasSecurityClass: {
                const hasSecurityClass = node.hasSecurityClass(message.securityClass);
                return { hasSecurityClass };
            }
            case command_1.NodeCommand.getHighestSecurityClass: {
                const highestSecurityClass = node.getHighestSecurityClass();
                return { highestSecurityClass };
            }
            case command_1.NodeCommand.testPowerlevel: {
                const framesAcked = await node.testPowerlevel(message.testNodeId, message.powerlevel, message.testFrameCount, (acknowledged, total) => {
                    clientsController.clients.forEach((client) => client.sendEvent({
                        source: "node",
                        event: "test powerlevel progress",
                        nodeId: message.nodeId,
                        acknowledged,
                        total,
                    }));
                });
                return { framesAcked };
            }
            case command_1.NodeCommand.checkLifelineHealth: {
                const summary = await node.checkLifelineHealth(message.rounds, (round, totalRounds, lastRating) => {
                    clientsController.clients.forEach((client) => client.sendEvent({
                        source: "node",
                        event: "check lifeline health progress",
                        nodeId: message.nodeId,
                        round,
                        totalRounds,
                        lastRating,
                    }));
                });
                return { summary };
            }
            case command_1.NodeCommand.checkRouteHealth: {
                const summary = await node.checkRouteHealth(message.targetNodeId, message.rounds, (round, totalRounds, lastRating) => {
                    clientsController.clients.forEach((client) => client.sendEvent({
                        source: "node",
                        event: "check route health progress",
                        nodeId: message.nodeId,
                        round,
                        totalRounds,
                        lastRating,
                    }));
                });
                return { summary };
            }
            case command_1.NodeCommand.getValue: {
                const value = node.getValue(message.valueId);
                return { value };
            }
            case command_1.NodeCommand.getEndpointCount: {
                const count = node.getEndpointCount();
                return { count };
            }
            case command_1.NodeCommand.interviewCC: {
                await node.interviewCC(message.commandClass);
                return {};
            }
            case command_1.NodeCommand.getState: {
                const state = (0, __1.dumpNode)(node, client.schemaVersion);
                return { state };
            }
            case command_1.NodeCommand.setKeepAwake: {
                node.keepAwake = message.keepAwake;
                return {};
            }
            case command_1.NodeCommand.setLocation: {
                node.location = message.location;
                if (((_c = message.updateCC) !== null && _c !== void 0 ? _c : true) &&
                    node.supportsCC(core_1.CommandClasses["Node Naming and Location"])) {
                    await node.commandClasses["Node Naming and Location"].setLocation(message.location);
                }
                return {};
            }
            case command_1.NodeCommand.setName: {
                node.name = message.name;
                if (((_d = message.updateCC) !== null && _d !== void 0 ? _d : true) &&
                    node.supportsCC(core_1.CommandClasses["Node Naming and Location"])) {
                    await node.commandClasses["Node Naming and Location"].setName(message.name);
                }
                return {};
            }
            case command_1.NodeCommand.getFirmwareUpdateProgress:
            case command_1.NodeCommand.isFirmwareUpdateInProgress: {
                const progress = node.isFirmwareUpdateInProgress();
                return { progress };
            }
            case command_1.NodeCommand.waitForWakeup: {
                await node.waitForWakeup();
                return {};
            }
            case command_1.NodeCommand.interview: {
                await node.interview();
                return {};
            }
            case command_1.NodeCommand.getValueTimestamp: {
                const timestamp = node.getValueTimestamp(message.valueId);
                return { timestamp };
            }
            case command_1.NodeCommand.manuallyIdleNotificationValue: {
                node.manuallyIdleNotificationValue(message.valueId);
                return {};
            }
            case command_1.NodeCommand.setDateAndTime: {
                const success = await node.setDateAndTime(message.date === undefined ? undefined : new Date(message.date));
                return { success };
            }
            default: {
                throw new error_1.UnknownCommandError(command);
            }
        }
    }
}
exports.NodeMessageHandler = NodeMessageHandler;
